@@ -403,9 +403,9 @@ sub convert_value_to_cpp($)
 {
   my ($text) = @_;
 
-  $$text =~ s"\bFALSE\b"<tt>false</tt>"g;
-  $$text =~ s"\bTRUE\b"<tt>true</tt>"g;
-  $$text =~ s"\bNULL\b"<tt>nullptr</tt>"g;
+  $$text =~ s"`?\bFALSE\b`?"<tt>false</tt>"g;
+  $$text =~ s"`?\bTRUE\b`?"<tt>true</tt>"g;
+  $$text =~ s"`?\bNULL\b`?"<tt>nullptr</tt>"g;
 
   # Enumerator names
   $$text =~ s/\b([A-Z]+)_([A-Z\d_]+)\b/&DocsParser::substitute_enumerator_name($1, $2)/eg;
@@ -423,6 +423,9 @@ sub remove_example_code($$)
   $example_removals +=
     ($$text =~ s"<programlisting>.*?</programlisting>"\n[C example ellipted]"sg);
   $example_removals += ($$text =~ s"\|\[.*?]\|"\n[C example ellipted]"sg);
+  # gi-docgen syntax.
+  # remove_example_code() is called after add_m4_quotes().
+  $example_removals += ($$text =~ s"(?:'__BT__`){3}[cC].*?(?:'__BT__`){3}"\n[C example ellipted]"sg);
 
   # See "MS Visual Studio" comment in gmmproc.in.
   print STDERR "gmmproc, $main::source, $obj_name: Example code discarded.\n"
@@ -472,6 +475,12 @@ sub append_parameter_docs($$;$)
   # due to imperfections in the C docs, and it's difficult to get the C docs
   # corrected, correct docs can be added to the docs_override.xml file.
 
+  if (scalar @docs_param_names != scalar @c_param_names)
+  {
+    print STDERR "DocsParser.pm: Warning, $$obj_function{name}\n" .
+      "  Incompatible parameter lists in the docs.xml file and the methods.defs file.\n";
+  }
+
   # Skip first param if this is a signal.
   if ($$obj_function{name} =~ /\w+::/)
   {
@@ -508,8 +517,14 @@ sub append_parameter_docs($$;$)
   # Skip the last param if it's an error output param.
   if (scalar @docs_param_names && $docs_param_names[-1] eq "error")
   {
-    pop(@docs_param_names);
-    pop(@c_param_names);
+    # If the number of parameters in @docs_param_names is not greater than
+    # the number of parameters in the _WRAP macro, the parameter called "error"
+    # is probably not an error output parameter.
+    if (!defined($objCppfunc) || scalar @docs_param_names > scalar @{$$objCppfunc{param_names}})
+    {
+      pop(@docs_param_names);
+      pop(@c_param_names);
+    }
   }
 
   my $cpp_param_names;
@@ -523,6 +538,11 @@ sub append_parameter_docs($$;$)
     {
       $out_param_index = $$param_mappings{OUT};
     }
+    if (scalar @docs_param_names != scalar @$cpp_param_names)
+    {
+      print STDERR "DocsParser.pm: Warning, $$obj_function{name}\n" .
+        "  Incompatible parameter lists in the docs.xml file and the _WRAP macro.\n";
+    }
   }
   my %param_name_mappings; # C name -> C++ name
 
@@ -635,6 +655,9 @@ sub remove_c_memory_handling_info($)
   # Don't modify the text, if $mem_funcs is part of example code.
   # remove_c_memory_handling_info() is called before remove_example_code().
   return if ($$text =~ m"(?:<informalexample>|<programlisting>|\|\[).*?$mem_funcs.*?(?:</informalexample>|</programlisting>|]\|)"s);
+  # gi-docgen syntax.
+  # remove_c_memory_handling_info() is called before add_m4_quotes().
+  return if ($$text =~ m"\`\`\`[cC].*?$mem_funcs.*?\`\`\`"s);
 
   # First try to remove the sentence containing $mem_funcs.
   # For simplicity, assume that a sentence is any string ending with a period.
@@ -686,6 +709,9 @@ sub convert_tags_to_doxygen($)
     # Also don't convert @enum and @var which are used for enum documentation.
     s" \@a (throws?|param|enum|var)\b" \@$1"g;
 
+    # gi-docgen uses [type@Module.*]. Don't convert them.
+    s"(\[[a-z]+?) \@a (.+?])"$1\@$2"g;
+
     s"^Note ?\d?: "\@note "mg;
     s"</?programlisting>""g;
     s"<!>""g;
@@ -721,9 +747,9 @@ sub convert_tags_to_doxygen($)
     # Doxygen is too dumb to handle &mdash;
     s"&mdash;" \@htmlonly&mdash;\@endhtmlonly "g;
 
-    s"\%?\bFALSE\b"<tt>false</tt>"g;
-    s"\%?\bTRUE\b"<tt>true</tt>"g;
-    s"\%?\bNULL\b"<tt>nullptr</tt>"g;
+    s"`?\%?\bFALSE\b`?"<tt>false</tt>"g;
+    s"`?\%?\bTRUE\b`?"<tt>true</tt>"g;
+    s"`?\%?\bNULL\b`?"<tt>nullptr</tt>"g;
 
     s"#?\bgboolean\b"<tt>bool</tt>"g;
     s"#?\bg(int|short|long)\b"<tt>$1</tt>"g;
@@ -806,32 +832,40 @@ sub substitute_identifiers($$)
   {
     # TODO: handle more than one namespace
 
-    # The gtk-doc syntax for links to symbols is described in the gtk-doc manual:
-    # yelp gtk-doc/help/manual/C/index.docbook
-    # then select "Documenting the code" and "Documenting symbols".
+    # gi-docgen syntax for links to symbols is described at
+    # https://gnome.pages.gitlab.gnome.org/gi-docgen/linking.html
 
     # Convert property names to C++.
     # The standard (and correct) gtk-doc way of referring to properties.
     s/(#[A-Z]\w+):([a-z\d-]+)/my $name = "$1::property_$2()"; $name =~ s"-"_"g; "$name";/ge;
-    # This is an incorrect format but widely used so correctly treat as a property.
+    # This is an incorrect format but widely used so correctly treat as a
+    # property.
     s/(\s)::([a-z\d-]+)(\s+property)/my $name = "$1property_$2()$3"; $name =~ s"-"_"g; "$name";/ge;
     # This one catches properties written in the gtk-doc block as for example
-    # '#GtkActivatable::related-action property'. The correct way to write it
+    # '#GtkActivatable::related-action property'.  The correct way to write it
     # would be 'GtkActivatable:related-action' (with a single colon and not
-    # two because the double colons are specifically for signals)
-    # but a few are written with the double colon in the gtk docs so this
+    # two because the double colons are specifically for signals -- see the
+    # gtk-doc docs:
+    # http://developer.gnome.org/gtk-doc-manual/unstable/documenting_symbols.html.en)
+    # but a few are written with the double colon in the gtk+ docs so this
     # protects against those errors.
     s/([A-Z]\w+)::([a-z\d-]+)(\s+property)/my $name = "$1::property_$2()$3"; $name =~ s"-"_"g; "$name";/ge;
+    # gi-docgen syntax.
+    s/\[(`?)property@(?:([A-Z]\w*)\.)?([A-Z]\w+):([a-z\d-]+)\1]/"$1" . &DocsParser::substitute_property_or_signal_name($doc_func, $2, $3, "property", $4) . "$1"/ge;
 
     # Convert signal names to C++.
     s/(^|\s)::([a-z\d-]+)(\(\))*([^:\w]|$)/my $name = "$1signal_$2()$4"; $name =~ s"-"_"g; "$name";/ge;
     s/(#[A-Z]\w+)::([a-z\d-]+)(\(\))*([^:\w]|$)/my $name = "$1::signal_$2()$4"; $name =~ s"-"_"g; "$name";/ge;
+    s/\[(`?)signal@(?:([A-Z]\w*)\.)?([A-Z]\w+)::([a-z\d-]+)\1]/"$1" . &DocsParser::substitute_property_or_signal_name($doc_func, $2, $3, "signal", $4) . "$1"/ge;
 
     # Type names
-    s/[#%]([A-Z][a-z]*)([A-Z][A-Za-z]+)\b/&DocsParser::substitute_type_name($1, $2)/eg;
+    s/[#%]([A-Z][a-z]*)([A-Z][A-Za-z]+)\b/&DocsParser::substitute_type_name($doc_func, $1, $2)/eg;
+    s/`([A-Z][a-z]*)([A-Z][A-Za-z]*[a-z])`/"`" . &DocsParser::substitute_type_name($doc_func, $1, $2) . "`"/eg;
+    s/\[(`?)(?:class|enum|error|flags|iface|struct|type)@(?:([A-Z]\w*)\.)?([A-Z]\w+)\1]/"$1" . &DocsParser::substitute_type_name($doc_func, $2, $3) . "$1"/eg;
 
     # Enumerator names
     s/[#%]([A-Z]+)_([A-Z\d_]+)\b/&DocsParser::substitute_enumerator_name($1, $2)/eg;
+    s/`([A-Z]+)_([A-Z\d_]+)`/"`" . &DocsParser::substitute_enumerator_name($1, $2) . "`"/eg;
 
     s/\bG:://g; #Rename G::Something to Something.
 
@@ -840,12 +874,18 @@ sub substitute_identifiers($$)
 
     # Replace C function names with C++ counterparts.
     s/\b([a-z]+_[a-z][a-z\d_]+) ?\(\)/&DocsParser::substitute_function($doc_func, $1)/eg;
+    s/\[(`?)id@([a-z\d_]+)\1]/"$1" . &DocsParser::substitute_function($doc_func, $2) . "$1"/eg;
+    s/\[(`?)(?:ctor|method)@(?:([A-Z]\w*)\.)?([A-Z]\w+)\.([a-z\d_]+)\1]/"$1" . &DocsParser::substitute_split_function($doc_func, $2, $3, $4) . "$1"/eg;
+    s/\[(`?)vfunc@(?:([A-Z]\w*)\.)?([A-Z]\w+)\.([a-z\d_]+)\1]/"$1" . &DocsParser::substitute_split_function($doc_func, $2, $3, $4 . "_vfunc") . "$1"/eg;
+    s/\[(`?)func@([\w.]+)\1]/"$1" . &DocsParser::substitute_func_function($doc_func, $2) . "$1"/eg;
   }
 }
 
-sub substitute_type_name($$)
+sub substitute_type_name($$$)
 {
-  my ($module, $name) = @_;
+  my ($doc_func, $module, $name) = @_;
+  $module = get_module_from_doc_func($doc_func) if !$module;
+
   my $c_name = $module . $name;
 
   if (exists $DocsParser::type_names{$c_name})
@@ -895,17 +935,91 @@ sub substitute_enumerator_name($$)
 
   # Don't apply the default substitution to these module names.
   # They are not really modules.
-  if (grep {$module eq $_} qw(HAS NO O SO AF))
+  if (grep {$module eq $_} qw(HAS NO O SO AF XDG))
   {
     return $c_name;
   }
 
   my $cxx_name = (($module eq "G") ? "" : (ucfirst(lc($module)) . "::")) . $name;
 
-  #print "DocsParser.pm: Assuming the enumerator $c_name shall become $cxx_name.\n";
+  print "DocsParser.pm: Assuming the enumerator $c_name shall become $cxx_name.\n";
+
   return $cxx_name;
 }
 
+sub substitute_property_or_signal_name($$$$$)
+{
+  # $doc_func can be the name of a property (ModuleClass:property_name),
+  # signal (ModuleClass::signal_name), function/method (module_class_method_name)
+  # or class/enum/etc. (ModuleClass).
+  my ($doc_func, $module, $class, $prop_or_sig, $name) = @_;
+  $module = get_module_from_doc_func($doc_func) if !$module;
+
+  my $prefix = $module . $class;
+  $name =~ s"-"_"g;
+  $name = $prop_or_sig . "_" . $name . "()";
+  if (index($doc_func, $prefix . ":") == 0)
+  {
+    # Documentation of property or signal in the same class as the referred
+    # property or signal.
+    return $name;
+  }
+  if (index($doc_func, ":") == -1)
+  {
+    # Documentation of a function or method.
+    if (my $defs_method = GtkDefs::lookup_method_dont_mark($doc_func))
+    {
+      if ($$defs_method{class} eq $prefix)
+      {
+        # Documentation of function/method in the same class as the referred
+        # property or signal.
+        return $name;
+      }
+    }
+  }
+  return $module . "::" . $class . "::" . $name;
+}
+
+sub substitute_split_function($$$$)
+{
+  my ($doc_func, $module, $class, $name) = @_;
+  $module = get_module_from_doc_func($doc_func) if !$module;
+
+  my $prefix = build_method_prefix($module, $class);
+
+  if ($doc_func =~ m/^$prefix/)
+  {
+    return $name . "()";
+  }
+  else
+  {
+    return $module . "::" . $class . "::" . $name . "()";
+  }
+}
+
+sub substitute_func_function($$)
+{
+  # $name == Gtk.WidgetPaintable.func or Gtk.func or func.
+  my ($doc_func, $name) = @_;
+  $name =~ s/\./_/g;
+  $name =~ s/([a-z])([A-Z])/$1_$2/g;
+  return DocsParser::substitute_function($doc_func, lc($name));
+}
+
+sub get_module_from_doc_func($)
+{
+  my ($doc_func) = @_;
+
+  if ($doc_func =~ /^([a-z]+)_/)
+  {
+    # Function name. gtk_foo_bar -> Gtk
+    return "\u$1";
+  }
+  # Class name. GtkFooBar -> Gtk
+  $doc_func =~ /^([A-Z][a-z]*?)[A-Z]/;
+  return $1;
+}
+
 sub substitute_function($$)
 {
   my ($doc_func, $name) = @_;
@@ -973,15 +1087,11 @@ sub lookup_object_of_method($$)
   return undef;
 }
 
-
 sub build_method_name($$$$)
 {
   my ($doc_func, $module, $class, $name) = @_;
 
-  my $prefix = $module . $class;
-
-  $prefix =~ s/([a-z])([A-Z])/$1_$2/g;
-  $prefix = lc($prefix) . '_';
+  my $prefix = build_method_prefix($module, $class);
 
   if($$name =~ m/^\Q$prefix\E/)
   {
@@ -992,5 +1102,15 @@ sub build_method_name($$$$)
   }
 }
 
+sub build_method_prefix($$)
+{
+  my ($module, $class) = @_;
+
+  my $prefix = $module . $class;
+
+  $prefix =~ s/([a-z])([A-Z])/$1_$2/g;
+  $prefix =~ s/^(Gdk|Gtk)_GL([A-Z][a-z])/$1_GL_$2/; # Special cases, add an underline
+  return lc($prefix) . '_';
+}
 
 1; # indicate proper module load.
